@@ -10,6 +10,8 @@
 
 #include "rubysocket.h"
 
+static VALUE sym_exception, sym_wait_writable;
+
 static VALUE sock_s_unpack_sockaddr_in(VALUE, VALUE);
 
 void
@@ -138,7 +140,6 @@ sock_initialize(int argc, VALUE *argv, VALUE sock)
     if (NIL_P(protocol))
         protocol = INT2FIX(0);
 
-    rb_secure(3);
     setup_domain_and_type(domain, &d, type, &t);
     fd = rsock_socket(d, t, NUM2INT(protocol));
     if (fd < 0) rb_sys_fail("socket(2)");
@@ -150,7 +151,7 @@ sock_initialize(int argc, VALUE *argv, VALUE sock)
 static VALUE
 io_call_close(VALUE io)
 {
-    return rb_funcall(io, rb_intern("close"), 0, 0);
+    return rb_funcallv(io, rb_intern("close"), 0, 0);
 }
 
 static VALUE
@@ -319,14 +320,14 @@ rsock_sock_s_socketpair(int argc, VALUE *argv, VALUE klass)
  * * +remote_sockaddr+ - the +struct+ sockaddr contained in a string or Addrinfo object
  *
  * === Example:
- * 	# Pull down Google's web page
- * 	require 'socket'
- * 	include Socket::Constants
- * 	socket = Socket.new( AF_INET, SOCK_STREAM, 0 )
- * 	sockaddr = Socket.pack_sockaddr_in( 80, 'www.google.com' )
- * 	socket.connect( sockaddr )
- * 	socket.write( "GET / HTTP/1.0\r\n\r\n" )
- * 	results = socket.read
+ *   # Pull down Google's web page
+ *   require 'socket'
+ *   include Socket::Constants
+ *   socket = Socket.new( AF_INET, SOCK_STREAM, 0 )
+ *   sockaddr = Socket.pack_sockaddr_in( 80, 'www.google.com' )
+ *   socket.connect( sockaddr )
+ *   socket.write( "GET / HTTP/1.0\r\n\r\n" )
+ *   results = socket.read
  *
  * === Unix-based Exceptions
  * On unix-based systems the following system exceptions may be raised if
@@ -440,7 +441,7 @@ sock_connect(VALUE sock, VALUE addr)
 
 /*
  * call-seq:
- *   socket.connect_nonblock(remote_sockaddr) => 0
+ *   socket.connect_nonblock(remote_sockaddr, [options]) => 0
  *
  * Requests a connection to be made on the given +remote_sockaddr+ after
  * O_NONBLOCK is set for the underlying file descriptor.
@@ -450,22 +451,22 @@ sock_connect(VALUE sock, VALUE addr)
  * * +remote_sockaddr+ - the +struct+ sockaddr contained in a string or Addrinfo object
  *
  * === Example:
- * 	# Pull down Google's web page
- * 	require 'socket'
- * 	include Socket::Constants
- * 	socket = Socket.new(AF_INET, SOCK_STREAM, 0)
- * 	sockaddr = Socket.sockaddr_in(80, 'www.google.com')
- * 	begin # emulate blocking connect
- * 	  socket.connect_nonblock(sockaddr)
- * 	rescue IO::WaitWritable
- * 	  IO.select(nil, [socket]) # wait 3-way handshake completion
- * 	  begin
- * 	    socket.connect_nonblock(sockaddr) # check connection failure
- * 	  rescue Errno::EISCONN
- * 	  end
- * 	end
- * 	socket.write("GET / HTTP/1.0\r\n\r\n")
- * 	results = socket.read
+ *   # Pull down Google's web page
+ *   require 'socket'
+ *   include Socket::Constants
+ *   socket = Socket.new(AF_INET, SOCK_STREAM, 0)
+ *   sockaddr = Socket.sockaddr_in(80, 'www.google.com')
+ *   begin # emulate blocking connect
+ *     socket.connect_nonblock(sockaddr)
+ *   rescue IO::WaitWritable
+ *     IO.select(nil, [socket]) # wait 3-way handshake completion
+ *     begin
+ *       socket.connect_nonblock(sockaddr) # check connection failure
+ *     rescue Errno::EISCONN
+ *     end
+ *   end
+ *   socket.write("GET / HTTP/1.0\r\n\r\n")
+ *   results = socket.read
  *
  * Refer to Socket#connect for the exceptions that may be thrown if the call
  * to _connect_nonblock_ fails.
@@ -477,24 +478,40 @@ sock_connect(VALUE sock, VALUE addr)
  * it is extended by IO::WaitWritable.
  * So IO::WaitWritable can be used to rescue the exceptions for retrying connect_nonblock.
  *
+ * By specifying `exception: false`, the options hash allows you to indicate
+ * that connect_nonblock should not raise an IO::WaitWritable exception, but
+ * return the symbol :wait_writable instead.
+ *
  * === See
  * * Socket#connect
  */
 static VALUE
-sock_connect_nonblock(VALUE sock, VALUE addr)
+sock_connect_nonblock(int argc, VALUE *argv, VALUE sock)
 {
+    VALUE addr;
+    VALUE opts = Qnil;
     VALUE rai;
     rb_io_t *fptr;
     int n;
 
+    rb_scan_args(argc, argv, "1:", &addr, &opts);
     SockAddrStringValueWithAddrinfo(addr, rai);
     addr = rb_str_new4(addr);
     GetOpenFile(sock, fptr);
     rb_io_set_nonblock(fptr);
     n = connect(fptr->fd, (struct sockaddr*)RSTRING_PTR(addr), RSTRING_SOCKLEN(addr));
     if (n < 0) {
-        if (errno == EINPROGRESS)
+        if (errno == EINPROGRESS) {
+           if (rsock_opt_false_p(opts, sym_exception)) {
+                return sym_wait_writable;
+            }
             rb_readwrite_sys_fail(RB_IO_WAIT_WRITABLE, "connect(2) would block");
+	}
+	if (errno == EISCONN) {
+           if (rsock_opt_false_p(opts, sym_exception)) {
+                return INT2FIX(0);
+            }
+	}
 	rsock_sys_fail_raddrinfo_or_sockaddr("connect(2)", addr, rai);
     }
 
@@ -511,18 +528,18 @@ sock_connect_nonblock(VALUE sock, VALUE addr)
  * * +local_sockaddr+ - the +struct+ sockaddr contained in a string or an Addrinfo object
  *
  * === Example
- * 	require 'socket'
+ *   require 'socket'
  *
- * 	# use Addrinfo
- * 	socket = Socket.new(:INET, :STREAM, 0)
- * 	socket.bind(Addrinfo.tcp("127.0.0.1", 2222))
- * 	p socket.local_address #=> #<Addrinfo: 127.0.0.1:2222 TCP>
+ *   # use Addrinfo
+ *   socket = Socket.new(:INET, :STREAM, 0)
+ *   socket.bind(Addrinfo.tcp("127.0.0.1", 2222))
+ *   p socket.local_address #=> #<Addrinfo: 127.0.0.1:2222 TCP>
  *
- * 	# use struct sockaddr
- * 	include Socket::Constants
- * 	socket = Socket.new( AF_INET, SOCK_STREAM, 0 )
- * 	sockaddr = Socket.pack_sockaddr_in( 2200, 'localhost' )
- * 	socket.bind( sockaddr )
+ *   # use struct sockaddr
+ *   include Socket::Constants
+ *   socket = Socket.new( AF_INET, SOCK_STREAM, 0 )
+ *   sockaddr = Socket.pack_sockaddr_in( 2200, 'localhost' )
+ *   socket.bind( sockaddr )
  *
  * === Unix-based Exceptions
  * On unix-based based systems the following system exceptions may be raised if
@@ -613,18 +630,18 @@ sock_bind(VALUE sock, VALUE addr)
  * * +backlog+ - the maximum length of the queue for pending connections.
  *
  * === Example 1
- * 	require 'socket'
- * 	include Socket::Constants
- * 	socket = Socket.new( AF_INET, SOCK_STREAM, 0 )
- * 	sockaddr = Socket.pack_sockaddr_in( 2200, 'localhost' )
- * 	socket.bind( sockaddr )
- * 	socket.listen( 5 )
+ *   require 'socket'
+ *   include Socket::Constants
+ *   socket = Socket.new( AF_INET, SOCK_STREAM, 0 )
+ *   sockaddr = Socket.pack_sockaddr_in( 2200, 'localhost' )
+ *   socket.bind( sockaddr )
+ *   socket.listen( 5 )
  *
  * === Example 2 (listening on an arbitrary port, unix-based systems only):
- * 	require 'socket'
- * 	include Socket::Constants
- * 	socket = Socket.new( AF_INET, SOCK_STREAM, 0 )
- * 	socket.listen( 1 )
+ *   require 'socket'
+ *   include Socket::Constants
+ *   socket = Socket.new( AF_INET, SOCK_STREAM, 0 )
+ *   socket.listen( 1 )
  *
  * === Unix-based Exceptions
  * On unix based systems the above will work because a new +sockaddr+ struct
@@ -700,27 +717,27 @@ rsock_sock_listen(VALUE sock, VALUE log)
  * * +flags+ - zero or more of the +MSG_+ options
  *
  * === Example
- * 	# In one file, start this first
- * 	require 'socket'
- * 	include Socket::Constants
- * 	socket = Socket.new( AF_INET, SOCK_STREAM, 0 )
- * 	sockaddr = Socket.pack_sockaddr_in( 2200, 'localhost' )
- * 	socket.bind( sockaddr )
- * 	socket.listen( 5 )
- * 	client, client_addrinfo = socket.accept
- * 	data = client.recvfrom( 20 )[0].chomp
- * 	puts "I only received 20 bytes '#{data}'"
- * 	sleep 1
- * 	socket.close
+ *   # In one file, start this first
+ *   require 'socket'
+ *   include Socket::Constants
+ *   socket = Socket.new( AF_INET, SOCK_STREAM, 0 )
+ *   sockaddr = Socket.pack_sockaddr_in( 2200, 'localhost' )
+ *   socket.bind( sockaddr )
+ *   socket.listen( 5 )
+ *   client, client_addrinfo = socket.accept
+ *   data = client.recvfrom( 20 )[0].chomp
+ *   puts "I only received 20 bytes '#{data}'"
+ *   sleep 1
+ *   socket.close
  *
- * 	# In another file, start this second
- * 	require 'socket'
- * 	include Socket::Constants
- * 	socket = Socket.new( AF_INET, SOCK_STREAM, 0 )
- * 	sockaddr = Socket.pack_sockaddr_in( 2200, 'localhost' )
- * 	socket.connect( sockaddr )
- * 	socket.puts "Watch this get cut short!"
- * 	socket.close
+ *   # In another file, start this second
+ *   require 'socket'
+ *   include Socket::Constants
+ *   socket = Socket.new( AF_INET, SOCK_STREAM, 0 )
+ *   sockaddr = Socket.pack_sockaddr_in( 2200, 'localhost' )
+ *   socket.connect( sockaddr )
+ *   socket.puts "Watch this get cut short!"
+ *   socket.close
  *
  * === Unix-based Exceptions
  * On unix-based based systems the following system exceptions may be raised if the
@@ -817,33 +834,33 @@ sock_recvfrom(int argc, VALUE *argv, VALUE sock)
  * * +flags+ - zero or more of the +MSG_+ options
  *
  * === Example
- * 	# In one file, start this first
- * 	require 'socket'
- * 	include Socket::Constants
- * 	socket = Socket.new(AF_INET, SOCK_STREAM, 0)
- * 	sockaddr = Socket.sockaddr_in(2200, 'localhost')
- * 	socket.bind(sockaddr)
- * 	socket.listen(5)
- * 	client, client_addrinfo = socket.accept
- * 	begin # emulate blocking recvfrom
- * 	  pair = client.recvfrom_nonblock(20)
- * 	rescue IO::WaitReadable
- * 	  IO.select([client])
- * 	  retry
- * 	end
- * 	data = pair[0].chomp
- * 	puts "I only received 20 bytes '#{data}'"
- * 	sleep 1
- * 	socket.close
+ *   # In one file, start this first
+ *   require 'socket'
+ *   include Socket::Constants
+ *   socket = Socket.new(AF_INET, SOCK_STREAM, 0)
+ *   sockaddr = Socket.sockaddr_in(2200, 'localhost')
+ *   socket.bind(sockaddr)
+ *   socket.listen(5)
+ *   client, client_addrinfo = socket.accept
+ *   begin # emulate blocking recvfrom
+ *     pair = client.recvfrom_nonblock(20)
+ *   rescue IO::WaitReadable
+ *     IO.select([client])
+ *     retry
+ *   end
+ *   data = pair[0].chomp
+ *   puts "I only received 20 bytes '#{data}'"
+ *   sleep 1
+ *   socket.close
  *
- * 	# In another file, start this second
- * 	require 'socket'
- * 	include Socket::Constants
- * 	socket = Socket.new(AF_INET, SOCK_STREAM, 0)
- * 	sockaddr = Socket.sockaddr_in(2200, 'localhost')
- * 	socket.connect(sockaddr)
- * 	socket.puts "Watch this get cut short!"
- * 	socket.close
+ *   # In another file, start this second
+ *   require 'socket'
+ *   include Socket::Constants
+ *   socket = Socket.new(AF_INET, SOCK_STREAM, 0)
+ *   sockaddr = Socket.sockaddr_in(2200, 'localhost')
+ *   socket.connect(sockaddr)
+ *   socket.puts "Watch this get cut short!"
+ *   socket.close
  *
  * Refer to Socket#recvfrom for the exceptions that may be thrown if the call
  * to _recvfrom_nonblock_ fails.
@@ -851,7 +868,7 @@ sock_recvfrom(int argc, VALUE *argv, VALUE sock)
  * Socket#recvfrom_nonblock may raise any error corresponding to recvfrom(2) failure,
  * including Errno::EWOULDBLOCK.
  *
- * If the exception is Errno::EWOULDBLOCK or Errno::AGAIN,
+ * If the exception is Errno::EWOULDBLOCK or Errno::EAGAIN,
  * it is extended by IO::WaitReadable.
  * So IO::WaitReadable can be used to rescue the exceptions for retrying recvfrom_nonblock.
  *
@@ -894,7 +911,7 @@ sock_accept(VALUE sock)
 
 /*
  * call-seq:
- *   socket.accept_nonblock => [client_socket, client_addrinfo]
+ *   socket.accept_nonblock([options]) => [client_socket, client_addrinfo]
  *
  * Accepts an incoming connection using accept(2) after
  * O_NONBLOCK is set for the underlying file descriptor.
@@ -903,32 +920,32 @@ sock_accept(VALUE sock)
  * and an Addrinfo, _client_addrinfo_.
  *
  * === Example
- * 	# In one script, start this first
- * 	require 'socket'
- * 	include Socket::Constants
- * 	socket = Socket.new(AF_INET, SOCK_STREAM, 0)
- * 	sockaddr = Socket.sockaddr_in(2200, 'localhost')
- * 	socket.bind(sockaddr)
- * 	socket.listen(5)
- * 	begin # emulate blocking accept
- * 	  client_socket, client_addrinfo = socket.accept_nonblock
- * 	rescue IO::WaitReadable, Errno::EINTR
- * 	  IO.select([socket])
- * 	  retry
- * 	end
- * 	puts "The client said, '#{client_socket.readline.chomp}'"
- * 	client_socket.puts "Hello from script one!"
- * 	socket.close
+ *   # In one script, start this first
+ *   require 'socket'
+ *   include Socket::Constants
+ *   socket = Socket.new(AF_INET, SOCK_STREAM, 0)
+ *   sockaddr = Socket.sockaddr_in(2200, 'localhost')
+ *   socket.bind(sockaddr)
+ *   socket.listen(5)
+ *   begin # emulate blocking accept
+ *     client_socket, client_addrinfo = socket.accept_nonblock
+ *   rescue IO::WaitReadable, Errno::EINTR
+ *     IO.select([socket])
+ *     retry
+ *   end
+ *   puts "The client said, '#{client_socket.readline.chomp}'"
+ *   client_socket.puts "Hello from script one!"
+ *   socket.close
  *
- * 	# In another script, start this second
- * 	require 'socket'
- * 	include Socket::Constants
- * 	socket = Socket.new(AF_INET, SOCK_STREAM, 0)
- * 	sockaddr = Socket.sockaddr_in(2200, 'localhost')
- * 	socket.connect(sockaddr)
- * 	socket.puts "Hello from script 2."
- * 	puts "The server said, '#{socket.readline.chomp}'"
- * 	socket.close
+ *   # In another script, start this second
+ *   require 'socket'
+ *   include Socket::Constants
+ *   socket = Socket.new(AF_INET, SOCK_STREAM, 0)
+ *   sockaddr = Socket.sockaddr_in(2200, 'localhost')
+ *   socket.connect(sockaddr)
+ *   socket.puts "Hello from script 2."
+ *   puts "The server said, '#{socket.readline.chomp}'"
+ *   socket.close
  *
  * Refer to Socket#accept for the exceptions that may be thrown if the call
  * to _accept_nonblock_ fails.
@@ -936,23 +953,31 @@ sock_accept(VALUE sock)
  * Socket#accept_nonblock may raise any error corresponding to accept(2) failure,
  * including Errno::EWOULDBLOCK.
  *
- * If the exception is Errno::EWOULDBLOCK, Errno::AGAIN, Errno::ECONNABORTED or Errno::EPROTO,
+ * If the exception is Errno::EWOULDBLOCK, Errno::EAGAIN, Errno::ECONNABORTED or Errno::EPROTO,
  * it is extended by IO::WaitReadable.
  * So IO::WaitReadable can be used to rescue the exceptions for retrying accept_nonblock.
  *
+ * By specifying `exception: false`, the options hash allows you to indicate
+ * that accept_nonblock should not raise an IO::WaitReadable exception, but
+ * return the symbol :wait_readable instead.
+ *
  * === See
  * * Socket#accept
  */
 static VALUE
-sock_accept_nonblock(VALUE sock)
+sock_accept_nonblock(int argc, VALUE *argv, VALUE sock)
 {
     rb_io_t *fptr;
     VALUE sock2;
     union_sockaddr buf;
+    struct sockaddr *addr = &buf.addr;
     socklen_t len = (socklen_t)sizeof buf;
 
     GetOpenFile(sock, fptr);
-    sock2 = rsock_s_accept_nonblock(rb_cSocket, fptr, &buf.addr, &len);
+    sock2 = rsock_s_accept_nonblock(argc, argv, rb_cSocket, fptr, addr, &len);
+
+    if (SYMBOL_P(sock2)) /* :wait_readable */
+	return sock2;
     return rb_assoc_new(sock2, rsock_io_socket_addrinfo(sock2, &buf.addr, len));
 }
 
@@ -965,28 +990,28 @@ sock_accept_nonblock(VALUE sock)
  * and an Addrinfo, _client_addrinfo_.
  *
  * === Example
- * 	# In one script, start this first
- * 	require 'socket'
- * 	include Socket::Constants
- * 	socket = Socket.new( AF_INET, SOCK_STREAM, 0 )
- * 	sockaddr = Socket.pack_sockaddr_in( 2200, 'localhost' )
- * 	socket.bind( sockaddr )
- * 	socket.listen( 5 )
- * 	client_fd, client_addrinfo = socket.sysaccept
- * 	client_socket = Socket.for_fd( client_fd )
- * 	puts "The client said, '#{client_socket.readline.chomp}'"
- * 	client_socket.puts "Hello from script one!"
- * 	socket.close
+ *   # In one script, start this first
+ *   require 'socket'
+ *   include Socket::Constants
+ *   socket = Socket.new( AF_INET, SOCK_STREAM, 0 )
+ *   sockaddr = Socket.pack_sockaddr_in( 2200, 'localhost' )
+ *   socket.bind( sockaddr )
+ *   socket.listen( 5 )
+ *   client_fd, client_addrinfo = socket.sysaccept
+ *   client_socket = Socket.for_fd( client_fd )
+ *   puts "The client said, '#{client_socket.readline.chomp}'"
+ *   client_socket.puts "Hello from script one!"
+ *   socket.close
  *
- * 	# In another script, start this second
- * 	require 'socket'
- * 	include Socket::Constants
- * 	socket = Socket.new( AF_INET, SOCK_STREAM, 0 )
- * 	sockaddr = Socket.pack_sockaddr_in( 2200, 'localhost' )
- * 	socket.connect( sockaddr )
- * 	socket.puts "Hello from script 2."
- * 	puts "The server said, '#{socket.readline.chomp}'"
- * 	socket.close
+ *   # In another script, start this second
+ *   require 'socket'
+ *   include Socket::Constants
+ *   socket = Socket.new( AF_INET, SOCK_STREAM, 0 )
+ *   sockaddr = Socket.pack_sockaddr_in( 2200, 'localhost' )
+ *   socket.connect( sockaddr )
+ *   socket.puts "Hello from script 2."
+ *   puts "The server said, '#{socket.readline.chomp}'"
+ *   socket.close
  *
  * Refer to Socket#accept for the exceptions that may be thrown if the call
  * to _sysaccept_ fails.
@@ -1033,7 +1058,6 @@ sock_gethostname(VALUE obj)
 
     char buf[RUBY_MAX_HOST_NAME_LEN+1];
 
-    rb_secure(3);
     if (gethostname(buf, (int)sizeof buf - 1) < 0)
 	rb_sys_fail("gethostname(3)");
 
@@ -1050,7 +1074,6 @@ sock_gethostname(VALUE obj)
 {
     struct utsname un;
 
-    rb_secure(3);
     uname(&un);
     return rb_str_new2(un.nodename);
 }
@@ -1072,7 +1095,7 @@ make_addrinfo(struct rb_addrinfo *res0, int norevlookup)
     for (res = res0->ai; res; res = res->ai_next) {
 	ary = rsock_ipaddr(res->ai_addr, res->ai_addrlen, norevlookup);
 	if (res->ai_canonname) {
-	    RARRAY_PTR(ary)[2] = rb_str_new2(res->ai_canonname);
+	    RARRAY_ASET(ary, 2, rb_str_new2(res->ai_canonname));
 	}
 	rb_ary_push(ary, INT2FIX(res->ai_family));
 	rb_ary_push(ary, INT2FIX(res->ai_socktype));
@@ -1117,8 +1140,9 @@ sock_sockaddr(struct sockaddr *addr, socklen_t len)
 static VALUE
 sock_s_gethostbyname(VALUE obj, VALUE host)
 {
-    rb_secure(3);
-    return rsock_make_hostent(host, rsock_addrinfo(host, Qnil, SOCK_STREAM, AI_CANONNAME), sock_sockaddr);
+    struct rb_addrinfo *res =
+	rsock_addrinfo(host, Qnil, AF_UNSPEC, SOCK_STREAM, AI_CANONNAME);
+    return rsock_make_hostent(host, res, sock_sockaddr);
 }
 
 /*
@@ -1382,16 +1406,16 @@ sock_s_getnameinfo(int argc, VALUE *argv)
 	sa = tmp;
 	MEMZERO(&hints, struct addrinfo, 1);
 	if (RARRAY_LEN(sa) == 3) {
-	    af = RARRAY_PTR(sa)[0];
-	    port = RARRAY_PTR(sa)[1];
-	    host = RARRAY_PTR(sa)[2];
+	    af = RARRAY_AREF(sa, 0);
+	    port = RARRAY_AREF(sa, 1);
+	    host = RARRAY_AREF(sa, 2);
 	}
 	else if (RARRAY_LEN(sa) >= 4) {
-	    af = RARRAY_PTR(sa)[0];
-	    port = RARRAY_PTR(sa)[1];
-	    host = RARRAY_PTR(sa)[3];
+	    af = RARRAY_AREF(sa, 0);
+	    port = RARRAY_AREF(sa, 1);
+	    host = RARRAY_AREF(sa, 3);
 	    if (NIL_P(host)) {
-		host = RARRAY_PTR(sa)[2];
+		host = RARRAY_AREF(sa, 2);
 	    }
 	    else {
 		/*
@@ -1496,7 +1520,7 @@ sock_s_getnameinfo(int argc, VALUE *argv)
 static VALUE
 sock_s_pack_sockaddr_in(VALUE self, VALUE port, VALUE host)
 {
-    struct rb_addrinfo *res = rsock_addrinfo(host, port, 0, 0);
+    struct rb_addrinfo *res = rsock_addrinfo(host, port, AF_UNSPEC, 0, 0);
     VALUE addr = rb_str_new((char*)res->ai->ai_addr, res->ai->ai_addrlen);
 
     rb_freeaddrinfo(res);
@@ -1767,7 +1791,7 @@ socket_s_ip_address_list(VALUE self)
     int ret;
     struct lifnum ln;
     struct lifconf lc;
-    char *reason = NULL;
+    const char *reason = NULL;
     int save_errno;
     int i;
     VALUE list = Qnil;
@@ -1828,7 +1852,7 @@ socket_s_ip_address_list(VALUE self)
     errno = save_errno;
 
     if (reason)
-	rb_sys_fail(reason);
+	rb_syserr_fail(save_errno, reason);
     return list;
 
 #elif defined(SIOCGIFCONF)
@@ -1913,7 +1937,7 @@ socket_s_ip_address_list(VALUE self)
     errno = save_errno;
 
     if (reason)
-	rb_sys_fail(reason);
+	rb_syserr_fail(save_errno, reason);
     return list;
 
 #undef EXTRA_SPACE
@@ -2025,7 +2049,7 @@ socket_s_ip_address_list(VALUE self)
 #endif
 
 void
-Init_socket()
+Init_socket(void)
 {
     rsock_init_basicsocket();
 
@@ -2150,11 +2174,11 @@ Init_socket()
 
     rb_define_method(rb_cSocket, "initialize", sock_initialize, -1);
     rb_define_method(rb_cSocket, "connect", sock_connect, 1);
-    rb_define_method(rb_cSocket, "connect_nonblock", sock_connect_nonblock, 1);
+    rb_define_method(rb_cSocket, "connect_nonblock", sock_connect_nonblock, -1);
     rb_define_method(rb_cSocket, "bind", sock_bind, 1);
     rb_define_method(rb_cSocket, "listen", rsock_sock_listen, 1);
     rb_define_method(rb_cSocket, "accept", sock_accept, 0);
-    rb_define_method(rb_cSocket, "accept_nonblock", sock_accept_nonblock, 0);
+    rb_define_method(rb_cSocket, "accept_nonblock", sock_accept_nonblock, -1);
     rb_define_method(rb_cSocket, "sysaccept", sock_sysaccept, 0);
 
     rb_define_method(rb_cSocket, "recvfrom", sock_recvfrom, -1);
@@ -2179,4 +2203,8 @@ Init_socket()
 #endif
 
     rb_define_singleton_method(rb_cSocket, "ip_address_list", socket_s_ip_address_list, 0);
+
+#undef rb_intern
+    sym_exception = ID2SYM(rb_intern("exception"));
+    sym_wait_writable = ID2SYM(rb_intern("wait_writable"));
 }
