@@ -98,11 +98,11 @@ module URI
     #
     def initialize(opts = {})
       @pattern = initialize_pattern(opts)
-      @pattern.each_value {|v| v.freeze}
+      @pattern.each_value(&:freeze)
       @pattern.freeze
 
       @regexp = initialize_regexp(@pattern)
-      @regexp.each_value {|v| v.freeze}
+      @regexp.each_value(&:freeze)
       @regexp.freeze
     end
 
@@ -401,7 +401,7 @@ module URI
       # host          = hostname | IPv4address | IPv6reference (RFC 2732)
       ret[:HOST] = host = "(?:#{hostname}|#{ipv4addr}|#{ipv6ref})"
       # port          = *digit
-      port = '\d*'
+      ret[:PORT] = port = '\d*'
       # hostport      = host [ ":" port ]
       ret[:HOSTPORT] = hostport = "#{host}(?::#{port})?"
 
