@@ -2,7 +2,7 @@
 
   ruby.c -
 
-  $Author: nobu $
+  $Author: ko1 $
   created at: Tue Aug 10 12:47:31 JST 1993
 
   Copyright (C) 1993-2007 Yukihiro Matsumoto
@@ -15,9 +15,8 @@
 #include <windows.h>
 #include <sys/cygwin.h>
 #endif
-#include "ruby/ruby.h"
-#include "ruby/encoding.h"
 #include "internal.h"
+#include "ruby/thread.h"
 #include "eval_intern.h"
 #include "dln.h"
 #include <stdio.h>
@@ -45,6 +44,9 @@
 #ifndef MAXPATHLEN
 # define MAXPATHLEN 1024
 #endif
+#ifndef O_ACCMODE
+# define O_ACCMODE (O_RDONLY | O_WRONLY | O_RDWR)
+#endif
 
 #include "ruby/util.h"
 
@@ -61,11 +63,14 @@ char *getenv();
 #define DEFAULT_RUBYGEMS_ENABLED "enabled"
 #endif
 
-#define DISABLE_BIT(bit) (1U << disable_##bit)
-enum disable_flag_bits {
-    disable_gems,
-    disable_rubyopt,
-    disable_flag_count
+#define FEATURE_BIT(bit) (1U << feature_##bit)
+enum feature_flag_bits {
+    feature_gems,
+    feature_did_you_mean,
+    feature_rubyopt,
+    feature_frozen_string_literal,
+    feature_frozen_string_literal_debug,
+    feature_flag_count
 };
 
 #define DUMP_BIT(bit) (1U << dump_##bit)
@@ -88,7 +93,7 @@ struct cmdline_options {
     int do_loop, do_print;
     int do_line, do_split;
     int do_search;
-    unsigned int disable;
+    unsigned int features;
     int verbose;
     int safe_level;
     unsigned int setids;
@@ -117,9 +122,12 @@ cmdline_options_init(struct cmdline_options *opt)
     opt->src.enc.index = src_encoding_index;
     opt->ext.enc.index = -1;
     opt->intern.enc.index = -1;
+    opt->features = ~0U;
 #if DISABLE_RUBYGEMS
-    opt->disable |= DISABLE_BIT(gems);
+    opt->features &= ~FEATURE_BIT(gems);
 #endif
+    opt->features &= ~FEATURE_BIT(frozen_string_literal);
+    opt->features &= ~FEATURE_BIT(frozen_string_literal_debug);
     return opt;
 }
 
@@ -132,6 +140,18 @@ static struct {
     char **argv;
 } origarg;
 
+static void
+show_usage_line(const char *str, unsigned int namelen, unsigned int secondlen, int help)
+{
+    const unsigned int w = 16;
+    const int wrap = help && namelen + secondlen - 2 > w;
+    printf("  %.*s%-*.*s%-*s%s\n", namelen-1, str,
+	   (wrap ? 0 : w - namelen + 1),
+	   (help ? secondlen-1 : 0), str + namelen,
+	   (wrap ? w + 3 : 0), (wrap ? "\n" : ""),
+	   str + namelen + secondlen);
+}
+
 static void
 usage(const char *name, int help)
 {
@@ -182,17 +202,13 @@ usage(const char *name, int help)
     };
     static const struct message features[] = {
 	M("gems",    "",        "rubygems (default: "DEFAULT_RUBYGEMS_ENABLED")"),
+	M("did_you_mean", "",   "did_you_mean (default: "DEFAULT_RUBYGEMS_ENABLED")"),
 	M("rubyopt", "",        "RUBYOPT environment variable (default: enabled)"),
+	M("frozen-string-literal", "", "freeze all string literals (default: disabled)"),
     };
-    int i, w = 16, num = numberof(usage_msg) - (help ? 1 : 0);
-#define SHOW(m) do { \
-	int wrap = help && (m).namelen + (m).secondlen - 2 > w; \
-	printf("  %.*s%-*.*s%-*s%s\n", (m).namelen-1, (m).str, \
-	       (wrap ? 0 : w - (m).namelen + 1), \
-	       (help ? (m).secondlen-1 : 0), (m).str + (m).namelen, \
-	       (wrap ? w + 3 : 0), (wrap ? "\n" : ""), \
-	       (m).str + (m).namelen + (m).secondlen); \
-    } while (0)
+    int i;
+    const int num = numberof(usage_msg) - (help ? 1 : 0);
+#define SHOW(m) show_usage_line((m).str, (m).namelen, (m).secondlen, help)
 
     printf("Usage: %s [switches] [--] [programfile] [arguments]\n", name);
     for (i = 0; i < num; ++i)
@@ -313,6 +329,7 @@ ruby_incpush_expand(const char *path)
     ruby_push_include(path, expand_include_path);
 }
 
+#undef UTF8_PATH
 #if defined _WIN32 || defined __CYGWIN__
 static HMODULE libruby;
 
@@ -329,6 +346,32 @@ rb_libruby_handle(void)
 {
     return libruby;
 }
+
+static inline void
+translit_char_bin(char *p, int from, int to)
+{
+    while (*p) {
+	if ((unsigned char)*p == from)
+	    *p = to;
+	p++;
+    }
+}
+
+# define UTF8_PATH 1
+#endif
+
+#ifndef UTF8_PATH
+# define UTF8_PATH 0
+#endif
+
+#if UTF8_PATH
+static VALUE
+str_conv_enc(VALUE str, rb_encoding *from, rb_encoding *to)
+{
+    return rb_str_conv_enc_opts(str, from, to,
+				ECONV_UNDEF_REPLACE|ECONV_INVALID_REPLACE,
+				Qnil);
+}
 #endif
 
 void ruby_init_loadpath_safe(int safe_level);
@@ -339,12 +382,38 @@ ruby_init_loadpath(void)
     ruby_init_loadpath_safe(0);
 }
 
+#if defined(LOAD_RELATIVE) && defined(HAVE_DLADDR)
+static VALUE
+dladdr_path(const void* addr)
+{
+    Dl_info dli;
+    VALUE fname, path;
+
+    if (!dladdr(addr, &dli)) {
+	return rb_str_new(0, 0);
+    }
+#ifdef __linux__
+    else if (dli.dli_fname == origarg.argv[0]) {
+	fname = rb_str_new_cstr("/proc/self/exe");
+	path = rb_readlink(fname, NULL);
+    }
+#endif
+    else {
+	fname = rb_str_new_cstr(dli.dli_fname);
+	path = rb_realpath_internal(Qnil, fname, 1);
+    }
+    rb_str_resize(fname, 0);
+    return path;
+}
+#endif
+
+#define INITIAL_LOAD_PATH_MARK rb_intern_const("@gem_prelude_index")
+
 void
 ruby_init_loadpath_safe(int safe_level)
 {
     VALUE load_path;
     ID id_initial_load_path_mark;
-    extern const char ruby_initial_load_paths[];
     const char *paths = ruby_initial_load_paths;
 #if defined LOAD_RELATIVE
 # if defined HAVE_DLADDR || defined __CYGWIN__ || defined _WIN32
@@ -365,20 +434,8 @@ ruby_init_loadpath_safe(int safe_level)
     sopath = rb_str_new(0, MAXPATHLEN);
     libpath = RSTRING_PTR(sopath);
     GetModuleFileName(libruby, libpath, MAXPATHLEN);
-#elif defined(__EMX__)
-    _execname(libpath, sizeof(libpath) - 1);
 #elif defined(HAVE_DLADDR)
-    Dl_info dli;
-    if (dladdr((void *)(VALUE)expand_include_path, &dli)) {
-	char fbuf[MAXPATHLEN];
-	char *f = dln_find_file_r(dli.dli_fname, getenv(PATH_ENV), fbuf, sizeof(fbuf));
-	VALUE fname = rb_str_new_cstr(f ? f : dli.dli_fname);
-	rb_str_freeze(fname);
-	sopath = rb_realpath_internal(Qnil, fname, 1);
-    }
-    else {
-	sopath = rb_str_new(0, 0);
-    }
+    sopath = dladdr_path((void *)(VALUE)expand_include_path);
     libpath = RSTRING_PTR(sopath);
 #endif
 
@@ -435,7 +492,9 @@ ruby_init_loadpath_safe(int safe_level)
 	    p = p2;
 	}
 #endif
+#if !VARIABLE_LIBPATH
 	*p = 0;
+#endif
     }
 #if !VARIABLE_LIBPATH
     else {
@@ -455,7 +514,6 @@ ruby_init_loadpath_safe(int safe_level)
 
 #define RUBY_RELATIVE(path, len) rb_str_buf_cat(BASEPATH(), (path), (len))
 #else
-    extern const char ruby_exec_prefix[];
     const size_t exec_prefix_len = strlen(ruby_exec_prefix);
 #define RUBY_RELATIVE(path, len) rubylib_path_new((path), (len))
 #define PREFIX_PATH() RUBY_RELATIVE(ruby_exec_prefix, exec_prefix_len)
@@ -466,7 +524,7 @@ ruby_init_loadpath_safe(int safe_level)
 	ruby_push_include(getenv("RUBYLIB"), identical_path);
     }
 
-    id_initial_load_path_mark = rb_intern_const("@gem_prelude_index");
+    id_initial_load_path_mark = INITIAL_LOAD_PATH_MARK;
     while (*paths) {
 	size_t len = strlen(paths);
 	VALUE path = RUBY_RELATIVE(paths, len);
@@ -642,8 +700,26 @@ moreswitches(const char *s, struct cmdline_options *opt, int envopt)
     rb_str_resize(argstr, 0);
 }
 
+static int
+name_match_p(const char *name, const char *str, size_t len)
+{
+    if (len == 0) return 0;
+    do {
+	while (TOLOWER(*str) == *name) {
+	    if (!--len || !*++str) return 1;
+	    ++name;
+	}
+	if (*str != '-' && *str != '_') return 0;
+	while (ISALNUM(*name)) name++;
+	if (*name != '-' && *name != '_') return 0;
+	++name;
+	++str;
+    } while (len > 0);
+    return !*name;
+}
+
 #define NAME_MATCH_P(name, str, len) \
-    ((len) < (int)sizeof(name) && strncmp((str), (name), (len)) == 0)
+    ((len) < (int)sizeof(name) && name_match_p((name), (str), (len)))
 
 #define UNSET_WHEN(name, bit, str, len)	\
     if (NAME_MATCH_P((name), (str), (len))) { \
@@ -660,11 +736,14 @@ moreswitches(const char *s, struct cmdline_options *opt, int envopt)
 static void
 enable_option(const char *str, int len, void *arg)
 {
-#define UNSET_WHEN_DISABLE(bit) UNSET_WHEN(#bit, DISABLE_BIT(bit), str, len)
-    UNSET_WHEN_DISABLE(gems);
-    UNSET_WHEN_DISABLE(rubyopt);
+#define SET_WHEN_ENABLE(bit) SET_WHEN(#bit, FEATURE_BIT(bit), str, len)
+    SET_WHEN_ENABLE(gems);
+    SET_WHEN_ENABLE(did_you_mean);
+    SET_WHEN_ENABLE(rubyopt);
+    SET_WHEN_ENABLE(frozen_string_literal);
+    SET_WHEN_ENABLE(frozen_string_literal_debug);
     if (NAME_MATCH_P("all", str, len)) {
-	*(unsigned int *)arg = 0U;
+	*(unsigned int *)arg = ~0U;
 	return;
     }
     rb_warn("unknown argument for --enable: `%.*s'", len, str);
@@ -673,11 +752,14 @@ enable_option(const char *str, int len, void *arg)
 static void
 disable_option(const char *str, int len, void *arg)
 {
-#define SET_WHEN_DISABLE(bit) SET_WHEN(#bit, DISABLE_BIT(bit), str, len)
-    SET_WHEN_DISABLE(gems);
-    SET_WHEN_DISABLE(rubyopt);
+#define UNSET_WHEN_DISABLE(bit) UNSET_WHEN(#bit, FEATURE_BIT(bit), str, len)
+    UNSET_WHEN_DISABLE(gems);
+    UNSET_WHEN_DISABLE(did_you_mean);
+    UNSET_WHEN_DISABLE(rubyopt);
+    UNSET_WHEN_DISABLE(frozen_string_literal);
+    UNSET_WHEN_DISABLE(frozen_string_literal_debug);
     if (NAME_MATCH_P("all", str, len)) {
-	*(unsigned int *)arg = ~0U;
+	*(unsigned int *)arg = 0U;
 	return;
     }
     rb_warn("unknown argument for --disable: `%.*s'", len, str);
@@ -1013,10 +1095,10 @@ proc_options(long argc, char **argv, struct cmdline_options *opt, int envopt)
                 ruby_verbose = Qtrue;
             }
 	    else if (is_option_with_arg("enable", Qtrue, Qtrue)) {
-		ruby_each_words(s, enable_option, &opt->disable);
+		ruby_each_words(s, enable_option, &opt->features);
 	    }
 	    else if (is_option_with_arg("disable", Qtrue, Qtrue)) {
-		ruby_each_words(s, disable_option, &opt->disable);
+		ruby_each_words(s, disable_option, &opt->features);
 	    }
 	    else if (is_option_with_arg("encoding", Qfalse, Qtrue)) {
 		char *p;
@@ -1251,15 +1333,12 @@ rb_f_chomp(int argc, VALUE *argv)
     return str;
 }
 
-/* blank function in dmyext.c or generated by enc/make_encmake.rb */
-extern void Init_enc(void);
-
 static VALUE
 process_options(int argc, char **argv, struct cmdline_options *opt)
 {
     NODE *tree = 0;
     VALUE parser;
-    VALUE iseq;
+    const rb_iseq_t *iseq;
     rb_encoding *enc, *lenc;
     const char *s;
     char fbuf[MAXPATHLEN];
@@ -1275,7 +1354,7 @@ process_options(int argc, char **argv, struct cmdline_options *opt)
 	return Qtrue;
     }
 
-    if (!(opt->disable & DISABLE_BIT(rubyopt)) &&
+    if ((opt->features & FEATURE_BIT(rubyopt)) &&
 	opt->safe_level == 0 && (s = getenv("RUBYOPT"))) {
 	VALUE src_enc_name = opt->src.enc.name;
 	VALUE ext_enc_name = opt->ext.enc.name;
@@ -1302,11 +1381,6 @@ process_options(int argc, char **argv, struct cmdline_options *opt)
 	ruby_show_copyright();
     }
 
-    if (opt->safe_level >= 4) {
-	OBJ_TAINT(rb_argv);
-	OBJ_TAINT(GET_VM()->load_path);
-    }
-
     if (!opt->e_script) {
 	if (argc == 0) {	/* no more args */
 	    if (opt->verbose)
@@ -1338,14 +1412,16 @@ process_options(int argc, char **argv, struct cmdline_options *opt)
 
     opt->script_name = rb_str_new_cstr(opt->script);
     opt->script = RSTRING_PTR(opt->script_name);
-#if defined DOSISH || defined __CYGWIN__
+
+#if _WIN32
+    translit_char_bin(RSTRING_PTR(opt->script_name), '\\', '/');
+#elif defined DOSISH
     translit_char(RSTRING_PTR(opt->script_name), '\\', '/');
 #endif
 
     ruby_gc_set_params(opt->safe_level);
     ruby_init_loadpath_safe(opt->safe_level);
     Init_enc();
-    rb_enc_find_index("encdb");
     lenc = rb_locale_encoding();
     rb_enc_associate(rb_progname, lenc);
     rb_obj_freeze(rb_progname);
@@ -1380,17 +1456,37 @@ process_options(int argc, char **argv, struct cmdline_options *opt)
     {
 	long i;
 	VALUE load_path = GET_VM()->load_path;
+	const ID id_initial_load_path_mark = INITIAL_LOAD_PATH_MARK;
 	for (i = 0; i < RARRAY_LEN(load_path); ++i) {
 	    VALUE path = RARRAY_AREF(load_path, i);
+	    int mark = rb_attr_get(path, id_initial_load_path_mark) == path;
 	    path = rb_enc_associate(rb_str_dup(path), lenc);
+	    if (mark) rb_ivar_set(path, id_initial_load_path_mark, path);
 	    RARRAY_ASET(load_path, i, path);
 	}
     }
     Init_ext();		/* load statically linked extensions before rubygems */
-    if (!(opt->disable & DISABLE_BIT(gems))) {
+    if (opt->features & FEATURE_BIT(gems)) {
 	rb_define_module("Gem");
     }
+    if (opt->features & FEATURE_BIT(did_you_mean)) {
+	rb_define_module("DidYouMean");
+    }
     ruby_init_prelude();
+    if (opt->features & FEATURE_BIT(frozen_string_literal)) {
+	VALUE option = rb_hash_new();
+	rb_hash_aset(option, ID2SYM(rb_intern_const("frozen_string_literal")), Qtrue);
+	rb_funcallv(rb_cISeq, rb_intern_const("compile_option="), 1, &option);
+    }
+    if (opt->features & FEATURE_BIT(frozen_string_literal_debug)) {
+	VALUE option = rb_hash_new();
+	rb_hash_aset(option, ID2SYM(rb_intern_const("frozen_string_literal_debug")), Qtrue);
+	rb_funcallv(rb_cISeq, rb_intern_const("compile_option="), 1, &option);
+    }
+#if UTF8_PATH
+    opt->script_name = str_conv_enc(opt->script_name, rb_utf8_encoding(), lenc);
+    opt->script = RSTRING_PTR(opt->script_name);
+#endif
     ruby_set_argv(argc, argv);
     process_sflag(&opt->sflag);
 
@@ -1415,8 +1511,10 @@ process_options(int argc, char **argv, struct cmdline_options *opt)
 	    eenc = lenc;
 	}
 	rb_enc_associate(opt->e_script, eenc);
-        ruby_set_script_name(opt->script_name);
-	require_libraries(&opt->req_list);
+	if (!(opt->dump & ~DUMP_BIT(version_v))) {
+	    ruby_set_script_name(opt->script_name);
+	    require_libraries(&opt->req_list);
+	}
         ruby_set_script_name(progname);
 
 	PREPARE_PARSE_MAIN({
@@ -1457,11 +1555,6 @@ process_options(int argc, char **argv, struct cmdline_options *opt)
     process_sflag(&opt->sflag);
     opt->xflag = 0;
 
-    if (opt->safe_level >= 4) {
-	FL_UNSET(rb_argv, FL_TAINT);
-	FL_UNSET(GET_VM()->load_path, FL_TAINT);
-    }
-
     if (opt->dump & DUMP_BIT(syntax)) {
 	printf("Syntax OK\n");
 	return Qtrue;
@@ -1497,7 +1590,7 @@ process_options(int argc, char **argv, struct cmdline_options *opt)
     });
 
     if (opt->dump & DUMP_BIT(insns)) {
-	rb_io_write(rb_stdout, rb_iseq_disasm(iseq));
+	rb_io_write(rb_stdout, rb_iseq_disasm((const rb_iseq_t *)iseq));
 	rb_io_flush(rb_stdout);
 	return Qtrue;
     }
@@ -1508,20 +1601,21 @@ process_options(int argc, char **argv, struct cmdline_options *opt)
 
     rb_set_safe_level(opt->safe_level);
 
-    return iseq;
+    return (VALUE)iseq;
 }
 
 struct load_file_arg {
     VALUE parser;
     VALUE fname;
     int script;
+    int xflag;
     struct cmdline_options *opt;
     VALUE f;
-    int xflag;
+    VALUE lineno;
 };
 
 static VALUE
-load_file_internal2(VALUE argp_v)
+load_file_internal(VALUE argp_v)
 {
     struct load_file_arg *argp = (struct load_file_arg *)argp_v;
     VALUE parser = argp->parser;
@@ -1535,6 +1629,7 @@ load_file_internal2(VALUE argp_v)
     ID set_encoding;
     int xflag = argp->xflag;
 
+    argp->script = 0;
     CONST_ID(set_encoding, "set_encoding");
     if (script) {
 	VALUE c = 1;		/* something not nil */
@@ -1612,8 +1707,10 @@ load_file_internal2(VALUE argp_v)
 	    if (f != rb_stdin) rb_io_close(f);
 	    f = Qnil;
 	}
-        ruby_set_script_name(opt->script_name);
-	require_libraries(&opt->req_list);	/* Why here? unnatural */
+	if (!(opt->dump & ~DUMP_BIT(version_v))) {
+	    ruby_set_script_name(opt->script_name);
+	    require_libraries(&opt->req_list);	/* Why here? unnatural */
+	}
     }
     if (opt->src.enc.index >= 0) {
 	enc = rb_enc_from_index(opt->src.enc.index);
@@ -1632,66 +1729,93 @@ load_file_internal2(VALUE argp_v)
     rb_funcall(f, set_encoding, 2, rb_enc_from_encoding(enc), rb_str_new_cstr("-"));
     tree = rb_parser_compile_file_path(parser, orig_fname, f, line_start);
     rb_funcall(f, set_encoding, 1, rb_parser_encoding(parser));
+    if (script && rb_parser_end_seen_p(parser)) argp->script = script;
     return (VALUE)tree;
 }
 
 static VALUE
-load_file_internal(VALUE arg)
+open_load_file(VALUE fname_v, int *xflag)
 {
-    extern VALUE rb_stdin;
-    struct load_file_arg *argp = (struct load_file_arg *)arg;
-    VALUE parser = argp->parser;
-    VALUE orig_fname = argp->fname;
-    VALUE fname_v = rb_str_encode_ospath(orig_fname);
     const char *fname = StringValueCStr(fname_v);
-    int script = argp->script;
     VALUE f;
-    NODE *tree;
-    int xflag = 0;
-    int state;
+    int e;
 
-    if (strcmp(fname, "-") == 0) {
+    if (RSTRING_LEN(fname_v) == 1 && fname[0] == '-') {
 	f = rb_stdin;
     }
     else {
-	int fd, mode = O_RDONLY;
+	int fd;
+	/* open(2) may block if fname is point to FIFO and it's empty. Let's
+	   use O_NONBLOCK. */
+#if defined O_NONBLOCK && HAVE_FCNTL && !(O_NONBLOCK & O_ACCMODE)
+	/* TODO: fix conflicting O_NONBLOCK in ruby/win32.h */
+# define MODE_TO_LOAD (O_RDONLY | O_NONBLOCK)
+#elif defined O_NDELAY && HAVE_FCNTL && !(O_NDELAY & O_ACCMODE)
+# define MODE_TO_LOAD (O_RDONLY | O_NDELAY)
+#else
+# define MODE_TO_LOAD (O_RDONLY)
+#endif
+	int mode = MODE_TO_LOAD;
 #if defined DOSISH || defined __CYGWIN__
 	{
 	    const char *ext = strrchr(fname, '.');
 	    if (ext && STRCASECMP(ext, ".exe") == 0) {
 		mode |= O_BINARY;
-		xflag = 1;
+		*xflag = 1;
 	    }
 	}
 #endif
+
 	if ((fd = rb_cloexec_open(fname, mode, 0)) < 0) {
 	    rb_load_fail(fname_v, strerror(errno));
 	}
-        rb_update_max_fd(fd);
-#if !defined DOSISH && !defined __CYGWIN__
+	rb_update_max_fd(fd);
+
+#if defined HAVE_FCNTL && MODE_TO_LOAD != O_RDONLY
+	/* disabling O_NONBLOCK */
+	if (fcntl(fd, F_SETFL, 0) < 0) {
+	    e = errno;
+	    (void)close(fd);
+	    rb_load_fail(fname_v, strerror(e));
+	}
+#endif
+
+#ifdef S_ISFIFO
 	{
 	    struct stat st;
-	    if (fstat(fd, &st) != 0)
-		rb_load_fail(fname_v, strerror(errno));
-	    if (S_ISDIR(st.st_mode)) {
-		errno = EISDIR;
-		rb_load_fail(fname_v, strerror(EISDIR));
+	    if (fstat(fd, &st) != 0) {
+		e = errno;
+		(void)close(fd);
+		rb_load_fail(fname_v, strerror(e));
+	    }
+	    if (S_ISFIFO(st.st_mode)) {
+		/*
+		  We need to wait if FIFO is empty. It's FIFO's semantics.
+		  rb_thread_wait_fd() release GVL. So, it's safe.
+		*/
+		rb_thread_wait_fd(fd);
 	    }
 	}
 #endif
+	if (!ruby_is_fd_loadable(fd)) {
+	    e = errno;
+	    (void)close(fd);
+	    rb_load_fail(fname_v, strerror(e));
+	}
+
 	f = rb_io_fdopen(fd, mode, fname);
     }
+    return f;
+}
 
-    argp->f = f;
-    argp->xflag = xflag;
-    tree = (NODE *)rb_protect(load_file_internal2, (VALUE)argp, &state);
-    if (state) {
-        if (f != rb_stdin)
-            rb_io_close(f);
-        rb_jump_tag(state);
-    }
+static VALUE
+restore_load_file(VALUE arg)
+{
+    struct load_file_arg *argp = (struct load_file_arg *)arg;
+    VALUE f = argp->f;
+    VALUE lineno = argp->lineno;
 
-    if (script && tree && rb_parser_end_seen_p(parser)) {
+    if (argp->script) {
 	/*
 	 * DATA is a File that contains the data section of the executed file.
 	 * To create a data section use <tt>__END__</tt>:
@@ -1709,12 +1833,6 @@ load_file_internal(VALUE arg)
     else if (f != rb_stdin) {
 	rb_io_close(f);
     }
-    return (VALUE)tree;
-}
-
-static VALUE
-restore_lineno(VALUE lineno)
-{
     return rb_gv_set("$.", lineno);
 }
 
@@ -1726,7 +1844,11 @@ load_file(VALUE parser, VALUE fname, int script, struct cmdline_options *opt)
     arg.fname = fname;
     arg.script = script;
     arg.opt = opt;
-    return (NODE *)rb_ensure(load_file_internal, (VALUE)&arg, restore_lineno, rb_gv_get("$."));
+    arg.xflag = 0;
+    arg.lineno = rb_gv_get("$.");
+    arg.f = open_load_file(rb_str_encode_ospath(fname), &arg.xflag);
+    return (NODE *)rb_ensure(load_file_internal, (VALUE)&arg,
+			     restore_load_file, (VALUE)&arg);
 }
 
 void *
@@ -1797,6 +1919,17 @@ set_arg0(VALUE val, ID id)
     rb_progname = rb_str_new_frozen(proc_setproctitle(rb_mProcess, val));
 }
 
+static inline VALUE
+external_str_new_cstr(const char *p)
+{
+#if UTF8_PATH
+    VALUE str = rb_utf8_str_new_cstr(p);
+    return str_conv_enc(str, NULL, rb_default_external_encoding());
+#else
+    return rb_external_str_new_cstr(p);
+#endif
+}
+
 /*! Sets the current script name to this value.
  *
  * This is similar to <code>$0 = name</code> in Ruby level but also affects
@@ -1806,7 +1939,7 @@ void
 ruby_script(const char *name)
 {
     if (name) {
-	rb_orig_progname = rb_progname = rb_external_str_new(name, strlen(name));
+	rb_orig_progname = rb_progname = external_str_new_cstr(name);
 	rb_vm_set_progname(rb_progname);
     }
 }
@@ -1913,7 +2046,7 @@ ruby_set_argv(int argc, char **argv)
 #endif
     rb_ary_clear(av);
     for (i = 0; i < argc; i++) {
-	VALUE arg = rb_external_str_new_cstr(argv[i]);
+	VALUE arg = external_str_new_cstr(argv[i]);
 
 	OBJ_FREEZE(arg);
 	rb_ary_push(av, arg);
@@ -1933,10 +2066,7 @@ ruby_process_options(int argc, char **argv)
     iseq = process_options(argc, argv, cmdline_options_init(&opt));
 
 #ifndef HAVE_SETPROCTITLE
-    {
-	extern void ruby_init_setproctitle(int argc, char *argv[]);
-	ruby_init_setproctitle(argc, argv);
-    }
+    ruby_init_setproctitle(argc, argv);
 #endif
 
     return (void*)(struct RData*)iseq;
