@@ -1,6 +1,5 @@
 begin
   require_relative 'dummyparser'
-  require_relative '../ruby/envutil'
   require 'test/unit'
   ripper_test = true
   module TestRipper; end
@@ -9,13 +8,10 @@ end
 
 class TestRipper::ParserEvents < Test::Unit::TestCase
 
-  # should be enabled
   def test_event_coverage
-    dispatched = Ripper::PARSER_EVENTS.map {|event,*| event }
-    dispatched.each do |e|
-      assert_equal true, respond_to?("test_#{e}", true),
-                   "event not tested: #{e.inspect}"
-    end
+    dispatched = Ripper::PARSER_EVENTS
+    tested = self.class.instance_methods(false).grep(/\Atest_(\w+)/) {$1.intern}
+    assert_empty dispatched-tested
   end
 
   def parse(str, nm = nil, &bl)
@@ -28,6 +24,14 @@ class TestRipper::ParserEvents < Test::Unit::TestCase
     parse(str, :compile_error) {|e, msg| return msg}
   end
 
+  def warning(str)
+    parse(str, :warning) {|e, *args| return args}
+  end
+
+  def warn(str)
+    parse(str, :warn) {|e, *args| return args}
+  end
+
   def test_program
     thru_program = false
     assert_equal '[void()]', parse('', :on_program) {thru_program = true}
@@ -249,9 +253,8 @@ class TestRipper::ParserEvents < Test::Unit::TestCase
     assert_equal true, thru_begin
   end
 
-  def test_binary
-    thru_binary = nil
-    %w"and or + - * / % ** | ^ & <=> > >= < <= == === != =~ !~ << >> && ||".each do |op|
+  %w"and or + - * / % ** | ^ & <=> > >= < <= == === != =~ !~ << >> && ||".each do |op|
+    define_method("test_binary(#{op})") do
       thru_binary = false
       parse("a #{op} b", :on_binary) {thru_binary = true}
       assert_equal true, thru_binary
@@ -351,12 +354,51 @@ class TestRipper::ParserEvents < Test::Unit::TestCase
     }
     assert_equal true, thru_call
     assert_equal "[call(ref(self),.,foo)]", tree
+
+    thru_call = false
+    assert_nothing_raised {
+      tree = parse("self.foo()", :on_call) {thru_call = true}
+    }
+    assert_equal true, thru_call
+    assert_equal "[call(ref(self),.,foo,[])]", tree
+
     thru_call = false
     assert_nothing_raised(bug2233) {
       tree = parse("foo.()", :on_call) {thru_call = true}
     }
     assert_equal true, thru_call
     assert_equal "[call(vcall(foo),.,call,[])]", tree
+
+    thru_call = false
+    assert_nothing_raised {
+      tree = parse("self::foo", :on_call) {thru_call = true}
+    }
+    assert_equal true, thru_call
+    assert_equal "[call(ref(self),::,foo)]", tree
+
+    thru_call = false
+    assert_nothing_raised {
+      tree = parse("self::foo()", :on_call) {thru_call = true}
+    }
+    assert_equal true, thru_call
+    assert_equal "[call(ref(self),::,foo,[])]", tree
+
+    thru_call = false
+    assert_nothing_raised(bug2233) {
+      tree = parse("foo::()", :on_call) {thru_call = true}
+    }
+    assert_equal true, thru_call
+    assert_equal "[call(vcall(foo),::,call,[])]", tree
+
+    thru_call = false
+    tree = parse("self&.foo", :on_call) {thru_call = true}
+    assert_equal true, thru_call
+    assert_equal "[call(ref(self),&.,foo)]", tree
+
+    thru_call = false
+    tree = parse("self&.foo()", :on_call) {thru_call = true}
+    assert_equal true, thru_call
+    assert_equal "[call(ref(self),&.,foo,[])]", tree
   end
 
   def test_excessed_comma
@@ -522,8 +564,13 @@ class TestRipper::ParserEvents < Test::Unit::TestCase
 
   def test_defs
     thru_defs = false
-    parse('def foo.bar; end', :on_defs) {thru_defs = true}
+    tree = parse('def foo.bar; end', :on_defs) {thru_defs = true}
     assert_equal true, thru_defs
+    assert_equal("[defs(vcall(foo),.,bar,[],bodystmt([void()]))]", tree)
+
+    thru_parse_error = false
+    tree = parse('def foo&.bar; end', :on_parse_error) {thru_parse_error = true}
+    assert_equal(true, thru_parse_error)
   end
 
   def test_do_block
@@ -548,6 +595,10 @@ class TestRipper::ParserEvents < Test::Unit::TestCase
     thru_dyna_symbol = false
     parse(':"#{foo}"', :on_dyna_symbol) {thru_dyna_symbol = true}
     assert_equal true, thru_dyna_symbol
+
+    thru_dyna_symbol = false
+    parse('{"#{foo}": 1}', :on_dyna_symbol) {thru_dyna_symbol = true}
+    assert_equal true, thru_dyna_symbol
   end
 
   def test_else
@@ -687,44 +738,67 @@ class TestRipper::ParserEvents < Test::Unit::TestCase
 
   def test_opassign
     thru_opassign = false
-    parse('a += b', :on_opassign) {thru_opassign = true}
+    tree = parse('a += b', :on_opassign) {thru_opassign = true}
+    assert_equal true, thru_opassign
+    assert_equal "[opassign(var_field(a),+=,vcall(b))]", tree
+    thru_opassign = false
+    tree = parse('a -= b', :on_opassign) {thru_opassign = true}
+    assert_equal true, thru_opassign
+    assert_equal "[opassign(var_field(a),-=,vcall(b))]", tree
+    thru_opassign = false
+    tree = parse('a *= b', :on_opassign) {thru_opassign = true}
+    assert_equal true, thru_opassign
+    assert_equal "[opassign(var_field(a),*=,vcall(b))]", tree
+    thru_opassign = false
+    tree = parse('a /= b', :on_opassign) {thru_opassign = true}
+    assert_equal true, thru_opassign
+    assert_equal "[opassign(var_field(a),/=,vcall(b))]", tree
+    thru_opassign = false
+    tree = parse('a %= b', :on_opassign) {thru_opassign = true}
+    assert_equal true, thru_opassign
+    assert_equal "[opassign(var_field(a),%=,vcall(b))]", tree
+    thru_opassign = false
+    tree = parse('a **= b', :on_opassign) {thru_opassign = true}
+    assert_equal true, thru_opassign
+    assert_equal "[opassign(var_field(a),**=,vcall(b))]", tree
+    thru_opassign = false
+    tree = parse('a &= b', :on_opassign) {thru_opassign = true}
+    assert_equal true, thru_opassign
+    assert_equal "[opassign(var_field(a),&=,vcall(b))]", tree
+    thru_opassign = false
+    tree = parse('a |= b', :on_opassign) {thru_opassign = true}
+    assert_equal "[opassign(var_field(a),|=,vcall(b))]", tree
     assert_equal true, thru_opassign
     thru_opassign = false
-    parse('a -= b', :on_opassign) {thru_opassign = true}
+    tree = parse('a <<= b', :on_opassign) {thru_opassign = true}
     assert_equal true, thru_opassign
+    assert_equal "[opassign(var_field(a),<<=,vcall(b))]", tree
     thru_opassign = false
-    parse('a *= b', :on_opassign) {thru_opassign = true}
+    tree = parse('a >>= b', :on_opassign) {thru_opassign = true}
     assert_equal true, thru_opassign
+    assert_equal "[opassign(var_field(a),>>=,vcall(b))]", tree
     thru_opassign = false
-    parse('a /= b', :on_opassign) {thru_opassign = true}
+    tree = parse('a &&= b', :on_opassign) {thru_opassign = true}
     assert_equal true, thru_opassign
+    assert_equal "[opassign(var_field(a),&&=,vcall(b))]", tree
     thru_opassign = false
-    parse('a %= b', :on_opassign) {thru_opassign = true}
+    tree = parse('a ||= b', :on_opassign) {thru_opassign = true}
     assert_equal true, thru_opassign
+    assert_equal "[opassign(var_field(a),||=,vcall(b))]", tree
     thru_opassign = false
-    parse('a **= b', :on_opassign) {thru_opassign = true}
+    tree = parse('a::X ||= c 1', :on_opassign) {thru_opassign = true}
     assert_equal true, thru_opassign
+    assert_equal "[opassign(const_path_field(vcall(a),X),||=,command(c,[1]))]", tree
+
     thru_opassign = false
-    parse('a &= b', :on_opassign) {thru_opassign = true}
+    tree = parse("self.foo += 1", :on_opassign) {thru_opassign = true}
     assert_equal true, thru_opassign
+    assert_equal "[opassign(field(ref(self),.,foo),+=,1)]", tree
+
     thru_opassign = false
-    parse('a |= b', :on_opassign) {thru_opassign = true}
-    assert_equal true, thru_opassign
-    thru_opassign = false
-    parse('a <<= b', :on_opassign) {thru_opassign = true}
-    assert_equal true, thru_opassign
-    thru_opassign = false
-    parse('a >>= b', :on_opassign) {thru_opassign = true}
-    assert_equal true, thru_opassign
-    thru_opassign = false
-    parse('a &&= b', :on_opassign) {thru_opassign = true}
-    assert_equal true, thru_opassign
-    thru_opassign = false
-    parse('a ||= b', :on_opassign) {thru_opassign = true}
-    assert_equal true, thru_opassign
-    thru_opassign = false
-    parse('a::X ||= c 1', :on_opassign) {thru_opassign = true}
+    tree = parse("self&.foo += 1", :on_opassign) {thru_opassign = true}
     assert_equal true, thru_opassign
+    assert_equal "[opassign(field(ref(self),&.,foo),+=,1)]", tree
   end
 
   def test_opassign_error
@@ -752,15 +826,31 @@ class TestRipper::ParserEvents < Test::Unit::TestCase
   end
 
   def test_params
+    arg = nil
     thru_params = false
-    parse('a {||}', :on_params) {thru_params = true}
+    parse('a {||}', :on_params) {|_, *v| thru_params = true; arg = v}
     assert_equal true, thru_params
+    assert_equal [nil, nil, nil, nil, nil, nil, nil], arg
     thru_params = false
-    parse('a {|x|}', :on_params) {thru_params = true}
+    parse('a {|x|}', :on_params) {|_, *v| thru_params = true; arg = v}
     assert_equal true, thru_params
+    assert_equal [["x"], nil, nil, nil, nil, nil, nil], arg
     thru_params = false
-    parse('a {|*x|}', :on_params) {thru_params = true}
+    parse('a {|*x|}', :on_params) {|_, *v| thru_params = true; arg = v}
     assert_equal true, thru_params
+    assert_equal [nil, nil, "*x", nil, nil, nil, nil], arg
+    thru_params = false
+    parse('a {|x: 1|}', :on_params) {|_, *v| thru_params = true; arg = v}
+    assert_equal true, thru_params
+    assert_equal [nil, nil, nil, nil, [["x:", "1"]], nil, nil], arg
+    thru_params = false
+    parse('a {|x:|}', :on_params) {|_, *v| thru_params = true; arg = v}
+    assert_equal true, thru_params
+    assert_equal [nil, nil, nil, nil, [["x:", false]], nil, nil], arg
+    thru_params = false
+    parse('a {|**x|}', :on_params) {|_, *v| thru_params = true; arg = v}
+    assert_equal true, thru_params
+    assert_equal [nil, nil, nil, nil, nil, "x", nil], arg
   end
 
   def test_paren
@@ -843,6 +933,14 @@ class TestRipper::ParserEvents < Test::Unit::TestCase
     assert_match(/rescue\(,var_field\(e\),\[2\]\)/, parsed)
   end
 
+  def test_rescue_class
+    thru_rescue = false
+    parsed = parse('begin; 1; rescue RuntimeError => e; 2; end', :on_rescue) {thru_rescue = true}
+    assert_equal true, thru_rescue
+    assert_match(/1.*rescue/, parsed)
+    assert_match(/rescue\(\[ref\(RuntimeError\)\],var_field\(e\),\[2\]\)/, parsed)
+  end
+
   def test_rescue_mod
     thru_rescue_mod = false
     parsed = parse('1 rescue 2', :on_rescue_mod) {thru_rescue_mod = true}
@@ -1196,14 +1294,30 @@ class TestRipper::ParserEvents < Test::Unit::TestCase
   def test_invalid_instance_variable_name
     assert_equal("`@1' is not allowed as an instance variable name", compile_error('@1'))
     assert_equal("`@%' is not allowed as an instance variable name", compile_error('@%'))
+    assert_equal("`@' without identifiers is not allowed as an instance variable name", compile_error('@'))
   end
 
   def test_invalid_class_variable_name
     assert_equal("`@@1' is not allowed as a class variable name", compile_error('@@1'))
     assert_equal("`@@%' is not allowed as a class variable name", compile_error('@@%'))
+    assert_equal("`@@' without identifiers is not allowed as a class variable name", compile_error('@@'))
   end
 
   def test_invalid_global_variable_name
     assert_equal("`$%' is not allowed as a global variable name", compile_error('$%'))
+    assert_equal("`$' without identifiers is not allowed as a global variable name", compile_error('$'))
+  end
+
+  def test_warning_shadowing
+    fmt, *args = warning("x = 1; tap {|;x|}")
+    assert_match(/shadowing outer local variable/, fmt)
+    assert_equal("x", args[0])
+    assert_match(/x/, fmt % args)
+  end
+
+  def test_warn_cr_in_middle
+    fmt = nil
+    assert_warn("") {fmt, *args = warn("\r;")}
+    assert_match(/encountered/, fmt)
   end
 end if ripper_test
